from ssh_ipykernel.status import Status
from ssh_ipykernel.kernel import SshKernel, SshKernelException
from ssh_ipykernel.utils import setup_logging

from databrickslabs_jupyterlab.local import get_db_config
from databrickslabs_jupyterlab.rest import Command, DatabricksApiException


class DatabricksKernelStatus(Status):
    def __init__(self, connection_info, logger, status_folder="~/.ssh_ipykernel"):
        super().__init__(connection_info, logger, status_folder)
        DatabricksKernelStatus.SPARK_RUNNING = 7
        DatabricksKernelStatus.MESSAGES[
            DatabricksKernelStatus.SPARK_RUNNING
        ] = "SparkSession created"

    def is_spark_running(self):
        """Check for Status.SPARK_RUNNING

        Returns:
            bool -- True if current status is Status.SPARK_RUNNING
        """
        return self._get_status() == DatabricksKernelStatus.SPARK_RUNNING

    def set_spark_running(self, pid=None, sudo=None):
        """Set current status to Status.SPARK_RUNNING
        """
        self._set_status(DatabricksKernelStatus.SPARK_RUNNING, pid, sudo)


class DatabricksKernel(SshKernel):
    def __init__(self, host, connection_info, python_path, sudo, timeout, env, no_spark):
        self._logger = setup_logging("DatabricksKernel")

        self.no_spark = no_spark

        self.dbjl_env = dict([e.split("=") for e in env[0].split(" ")])
        self._logger.debug("Environment = %s", self.dbjl_env)
        self.profile = self.dbjl_env.get("DBJL_PROFILE", None)
        self.host, self.token = get_db_config(self.profile)

        self.cluster_id = self.dbjl_env.get("DBJL_CLUSTER", None)

        self.python_command = None
        self.scala_context_id = None
        if not no_spark:
            # create remote executions context and retrieve its python path
            python_path = self.create_execution_context()
        self._logger.info("Remote python path: %s", python_path)

        super().__init__(
            host,
            connection_info,
            python_path,
            sudo=sudo,
            timeout=timeout,
            env=env,
            logger=self._logger,
        )

        self.kernel_status = DatabricksKernelStatus(connection_info, self._logger)

    def create_execution_context(self):
        if self.profile is None:
            self._logger.error("Environment variable DBJL_PROFILE is not set")
            self.no_spark = True
            return None

        if self.cluster_id is None:
            self._logger.error("Environment variable DBJL_CLUSTER is not set")
            self.no_spark = True
            return None

        self._logger.debug("Create Execution Context")

        self._logger.debug(
            "profile=%s, host=%s, cluster_id=%s", self.profile, self.host, self.cluster_id
        )

        # Create remote Python Context
        try:
            self.python_command = Command(
                url=self.host, cluster_id=self.cluster_id, token=self.token
            )
            self._logger.debug("Created remote Python context %s", self.python_command.context.id)
        except DatabricksApiException as ex:
            self._logger.error(str(ex))
            raise SshKernelException("Cannot create python execution context on remote cluster")

        # Create remote Scala Context
        try:
            self.scala_command = Command(
                url=self.host, cluster_id=self.cluster_id, token=self.token, language="scala"
            )
            self.scala_context_id = self.scala_command.context.id
            self._logger.debug("Created remote Scala context %s", self.scala_context_id)
        except DatabricksApiException as ex:
            self._logger.error(str(ex))
            self._logger.error("Cannot create Scala context, %%scala will not be possible")

        self._logger.info("Gateway created for cluster '%s'", self.cluster_id)

        try:
            cmd = "import sys; print(sys.executable.split('/bin/')[0])"
            result = self.python_command.execute(cmd)
        except Exception as ex:  # pylint: disable=broad-except
            self._logger.error("ERROR %s: %s", type(ex), ex)
            raise SshKernelException("Cannot retrieve python executable from remote cluster")

        return result[1]

    def kernel_customize(self):
        if self.no_spark:
            self._logger.info("This kernel will have no Spark Session (--no-spark)")
            return None

        self._logger.debug("Create Spark Session")

        # Fetch auth_token and gateway port ...
        #
        try:
            cmd = (
                "c=sc._gateway.client.gateway_client; "
                + 'print(c.gateway_parameters.auth_token + "|" + str(c.port))'
            )
            result = self.python_command.execute(cmd)
        except Exception as ex:  # pylint: disable=broad-except
            self._logger.error("ERROR %s: %s", type(ex), ex)
            raise SshKernelException("Cannot retrieve py4j gateway from remote cluster")

        gw_token, gw_port = result[1].split("|")
        gw_port = int(gw_port)
        self._logger.debug("Gateway token=%s, port=%s", gw_token, gw_port)

        cmd = (
            "from databrickslabs_jupyterlab.connect import dbcontext; "
            + "dbcontext(progressbar=True"
            + ", gw_port={}, gw_token='{}', token='{}', scala_context_id='{}')".format(
                gw_port, gw_token, self.token, self.scala_context_id
            )
        )
        try:
            result = self.kc.execute_interactive(
                cmd, silent=True, store_history=False, user_expressions={"spark": "spark.version"}
            )
            if result["content"]["status"] != "error":
                self.kernel_status.set_spark_running()
            else:
                self._logger.error("Error: Cluster unreachable")
                self.kernel_status.set_unreachable()
                raise SshKernelException("Cannot create SparkSession on remote cluster")

        except Exception as ex:  # pylint: disable=broad-except
            self._logger.error("ERROR %s: %s", type(ex), ex)
            self.kernel_status.set_connect_failed()
            raise SshKernelException("Cannot access kernel on remote cluster")

    def close(self):
        if not self.no_spark:
            try:
                self.python_command.close()
                self._logger.info("Remote Python context closed")
            except Exception as ex:  # pylint: disable=broad-except
                self._logger.error("%s: Failed closing Remote Python context: %s", type(ex), ex)
            try:
                self.scala_command.close()
                self._logger.info("Remote Scala context closed")
            except Exception as ex:  # pylint: disable=broad-except
                self._logger.error("%s: Failed closing Remote Scala context: %s", type(ex), ex)

        super().close()



import os
import sys

from IPython import get_ipython


class NotebookExit(BaseException):
    pass


class Notebook:
    def __init__(self):
        self.ipython = get_ipython()
        self._showtraceback = self.ipython._showtraceback
        self.ipython._showtraceback = self._exception_handler

    def run(self, notebook):
        path, ext = os.path.splitext(notebook)
        if not ext in [".ipynb", ".py", ""]:
            print("Only notebooks with ending .ipynb or python files with ending .py supported")
        else:
            if path.startswith("dbfs:"):
                path = path[5:]
            path = "/dbfs/" + path.strip("/") + ext
            if ext == "":
                if os.path.exists(path + ".ipynb"):
                    path = path + ".ipynb"
                elif os.path.exists(path + ".py"):
                    path = path + ".py"
            if os.path.exists(path):
                print("Running '{}'".format(path))
                return self.ipython.magic("run %s" % path)
            else:
                print("Error, '{}' does not exist".format(path))

    def _exception_handler(self, exception_type, exception, traceback):
        if exception_type == NotebookExit:
            print("Notebook exited: %s" % (exception), file=sys.stderr)
        else:
            self._showtraceback(exception_type, exception, traceback)

    def exit(self, msg):
        raise NotebookExit(msg)



import os
from notebook.utils import url_path_join

from databrickslabs_jupyterlab._version import __version__, __version_info__


def is_remote():
    """Check whether the current context is on the remote cluster
    
    Returns:
        bool: True if remote else False
    """
    return os.environ.get("DBJL_HOST", None) is not None


def is_azure():
    """Check whether the current context is on Azure Databricks
    
    Returns:
        bool: True if Azure Databricks else False
    """
    return os.environ.get("DBJL_ORG", None) is not None


def load_jupyter_server_extension(nbapp):
    """
    Called during notebook start
    """
    if not is_remote():
        from databrickslabs_jupyterlab.status import KernelHandler, DbStartHandler, DbStatusHandler

        KernelHandler.nbapp = nbapp
        base_url = nbapp.web_app.settings["base_url"]
        status_route = url_path_join(base_url, "/databrickslabs-jupyterlab-status")
        start_route = url_path_join(base_url, "/databrickslabs-jupyterlab-start")
        nbapp.web_app.add_handlers(
            ".*", [(status_route, DbStatusHandler), (start_route, DbStartHandler)]
        )



import json
import os
import re
import tempfile

import databrickslabs_jupyterlab
from databrickslabs_jupyterlab.remote import get_remote_packages
from databrickslabs_jupyterlab.local import print_ok, print_error
from databrickslabs_jupyterlab.utils import bye, execute_script


WHITELIST = [
    "boto3",
    "jsonschema",
    "hyperopt",
    "keras-applications",
    "keras-preprocessing",
    "keras",
    "lxml",
    "matplotlib",
    "mleap",
    "mlflow",
    "numba",
    "numpy",
    "pandas",
    "patsy",
    "pillow",
    "pyarrow",
    "python-dateutil",
    "pyparsing",
    "scikit-learn",
    "scipy",
    "seaborn",
    "simplejson",
    "statsmodels",
    "tabulate",
    "tensorboard",
    "tensorboardx",
    "tensorflow-estimator",
    "tensorflow",
    "torch",
    "torchvision",
    "urllib3",
]

BLACKLIST = [
    "absl-py",
    "ansi2html",
    "asn1crypto",
    "astor",
    "attrs",
    "backcall",
    "backports.shutil-get-terminal-size",
    "bcrypt",
    "bleach",
    "blessings",
    "brewer2mpl",
    "certifi",
    "cffi",
    "chardet",
    "click",
    "colorama",
    "configobj",
    "configparser",
    "cryptography",
    "cycler",
    "cython",
    "databricks-cli",
    "databrickslabs-jupyterlab",
    "decorator",
    "defusedxml",
    "docker",
    "docopt",
    "docutils",
    "entrypoints",
    "enum34",
    "et-xmlfile",
    "flask",
    "freetype-py",
    "funcsigs",
    "fusepy",
    "future",
    "gast",
    "gitdb2",
    "gitpython",
    "grpcio",
    "gunicorn",
    "h5py",
    "horovod",
    "html5lib",
    "idna",
    "inquirer",
    "ipaddress",
    "ipykernel",
    "ipython",
    "ipython-genutils",
    "ipywidgets",
    "itsdangerous",
    "jdcal",
    "jedi",
    "jinja2",
    "jmespath",
    "jsonschema",
    "jupyter-client",
    "jupyter-core",
    "llvmlite",
    "mako",
    "markdown",
    "markupsafe",
    "mistune",
    "mkl-fft",
    "mkl-random",
    "mock",
    "msgpack",
    "msgpack-python",
    "nbconvert",
    "nbformat",
    "ndg-httpsclient",
    "networkx",
    "nose",
    "nose-exclude",
    "notebook",
    "olefile",
    "openpyxl",
    "openpyxl",
    "packaging",
    "pandocfilters",
    "paramiko",
    "parso",
    "pathlib2",
    "pexpect",
    "pickleshare",
    "pip",
    "ply",
    "prometheus-client",
    "prompt-toolkit",
    "protobuf",
    "psutil",
    "psycopg2",
    "ptyprocess",
    "py4j",
    "pyasn1",
    "pycparser",
    "pycurl",
    "pygments",
    "pygobject",
    "pymongo",
    "pynacl",
    "pypng",
    "pysocks",
    "python-apt",
    "python-dateutil",
    "python-editor",
    "python-geohash",
    "pyopenssl",
    "pytz",
    "pyyaml",
    "pyzmq",
    "querystring-parser",
    "readchar",
    "requests",
    "scour",
    "send2trash",
    "setuptools",
    "simplegeneric",
    "singledispatch",
    "six",
    "smmap2",
    "sqlparse",
    "ssh-config",
    "ssh-import-id",
    "tabulate",
    "termcolor",
    "terminado",
    "testpath",
    "texttable",
    "tornado",
    "tqdm",
    "traitlets",
    "unattended-upgrades",
    "version-parser",
    "virtualenv",
    "wcwidth",
    "webencodings",
    "websocket-client",
    "werkzeug",
    "wheel",
    "widgetsnbextension",
    "wrapt",
]


def _set_conda_env(env_name):
    if env_name is None:
        return ""
    else:
        return "source $(conda info | awk '/base env/ {print $4}')/bin/activate '%s'\n" % env_name


def validate(env_name=None, labext=False):
    script = _set_conda_env(env_name)

    if labext:
        script += "jupyter-labextension list"
    else:
        script += """
            pip show databrickslabs_jupyterlab
            jupyter-serverextension list
        """

    print("* Validating conda environment")
    execute_script(
        script,
        "Successfully validated conda environment",
        "Error while validating conda environment",
    )


# def update_env(env_file):
#     script = "conda env update --file %s" % (env_file)

#     print("* Updating conda environment")
#     execute_script(
#         script, "Successfully updated conda environment", "Error while updating conda environment",
#     )
#     validate()


def install_env(env_file, env_name):
    script = "conda env create -n %s -f %s" % (env_name, env_file,)

    print("* Installing conda environment %s" % env_name)
    execute_script(
        script,
        "Successfully installed conda environment",
        "Error while installing conda environment",
    )
    validate(env_name)


# def install_labextensions(labext_file, env_name=None):
#     with open(labext_file, "r") as fd:
#         extensions = " ".join(fd.read().split("\n"))
#     script = _set_conda_env(env_name)
#     script += "jupyter labextension install --no-build %s\n" % extensions
#     script += "jupyter lab build --dev-build=True --minimize=False\n"

#     print("* Installing jupyterlab extensions")
#     execute_script(
#         script,
#         "Successfully installed jupyter labextensions",
#         "Error while installing jupyter labextensions",
#     )
#     validate(env_name, labext=True)


# def update_local():
#     module_path = os.path.dirname(databrickslabs_jupyterlab.__file__)

#     env_file = os.path.join(module_path, "lib", "env.yml")
#     update_env(env_file)

#     labext_file = os.path.join(module_path, "lib", "labextensions.txt")
#     install_labextensions(labext_file)
#     usage(os.environ.get("CONDA_DEFAULT_ENV", "unknown"))


def install(profile, host, token, cluster_id, cluster_name, use_whitelist):
    print("\n* Installation of local environment to mirror a remote Databricks cluster")
    result = get_remote_packages(cluster_id, host, token)
    if result[0] != 0:
        print_error(result[1])
        bye(1)
    libs = json.loads(result[1])

    if use_whitelist:
        print_ok("   => Using whitelist to select packages")
        ds_libs = [lib for lib in libs if lib["name"].lower() in WHITELIST]
    else:
        print_ok("   => Using blacklist to select packages")
        ds_libs = [lib for lib in libs if lib["name"].lower() not in BLACKLIST]

    ds_yml = ""
    for lib in ds_libs:
        if lib["name"] == "python":  # just artificially added
            python_version = lib["version"]
        else:
            if lib["name"] in ["hyperopt", "torchvision"]:
                r = re.compile(r"(\d+\.\d+.\d+)(.*)")
                version = r.match(lib["version"]).groups()[0]
            elif lib["name"] in ["tensorboardx"]:
                r = re.compile(r"(\d+\.\d+)(.*)")
                version = r.match(lib["version"]).groups()[0]
            else:
                version = lib["version"]
            ds_yml += "    - %s==%s\n" % (lib["name"], version)

    module_path = os.path.dirname(databrickslabs_jupyterlab.__file__)
    env_file = os.path.join(module_path, "lib", "env.yml")
    with open(env_file, "r") as fd:
        master_yml = fd.read()
    lines = master_yml.split("\n")
    for i in range(len(lines)):
        if lines[i].startswith("dependencies"):
            lines.insert(i + 1, "  - python=%s" % python_version)
            break
    master_yml = "\n".join(lines)

    print("\n Installed environment \n")
    print(master_yml)
    print("\n    # Data Science Libs\n")
    print(ds_yml + "\n")

    with tempfile.TemporaryDirectory() as tmpdir:
        env_file = os.path.join(tmpdir, "env.yml")
        with open(env_file, "w") as fd:
            fd.write(master_yml)
            fd.write("\n    # Data Science Libs\n")
            fd.write(ds_yml)
            fd.write("\n")

        env_name = cluster_name.replace(" ", "_")
        answer = input("    => Provide a conda environment name (default = %s): " % env_name)
        if answer != "":
            env_name = answer.replace(" ", "_")

        install_env(env_file, env_name)

        labext_file = os.path.join(module_path, "lib", "labextensions.txt")
        install_labextensions(labext_file, env_name)

    usage(env_name)


def usage(env_name):
    print_ok(
        """
= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 
                              Q U I C K S T A R T
= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 

Prerequisites:
--------------

- Install Databricks CLI and configure profile(s) for your cluster(s)

  AWS: https://docs.databricks.com/user-guide/dev-tools/databricks-cli.html)
  Azure: https://docs.azuredatabricks.net/user-guide/dev-tools/databricks-cli.html

- Create an ssh key pair called ~/.ssh/id_<profile> for each cluster and 
  add the public key to the cluster SSH configuration


databrickslabs-jupyterlab:
--------------------------

1) Show help

    conda activate %s
    databrickslabs-jupyterlab -h

2) Create jupyter kernel for remote cluster

    Databricks on AWS:
        databrickslabs-jupyterlab <profile> -k [-i cluster-id]
    
    Azure Databricks:
        databrickslabs-jupyterlab <profile> -k -o <organisation>

3) Compare local and remote python package versions
    
    databrickslabs-jupyterlab <profile> -v all|same|diff

4) Start jupyter lab to use the kernel(s) created in 2)

    databrickslabs-jupyterlab <profile> -l [-i cluster-id]


5) Check currently available profiles 

    databrickslabs-jupyterlab -p

6) Download a demo notebook from docs.databricks.com (experimental)

    databrickslabs-jupyterlab -n https://docs.databricks.com/_static/notebooks/delta/xyz.html


"""
        % env_name
    )



from ipywidgets import Select, HBox, VBox, Button, Output
from IPython.display import display, HTML
import pandas as pd


def _px(val):
    return "%dpx" % val


class DBList:
    def __init__(self, spark, console):
        self.spark = spark
        self.console = console
        self.height = 200
        self.db = None
        self.tbl = None
        self.list = None
        self.refresh = None

    def _set_status(self, status):
        with self.console:
            print(status, "...")

    def _clear_status(self):
        self.console.clear_output()

    def create(self, width=120, height=200):
        self.height = height
        self.refresh = Button(icon="refresh", layout={"width": "40px"})
        self.refresh.on_click(self.on_refresh)
        self.list = Select(
            options=[], disabled=False, layout={"width": _px(width), "height": _px(height)}
        )
        self.list.observe(self.on_click, names="value")
        self.update()

    def on_refresh(self, _):
        self.update()

    def on_click(self, change):
        self.console.clear_output()
        if change["old"] is not None:
            selected = change["new"]
            self.update(selected)

    def update(self, selected=None):
        pass


class DatabaseList(DBList):
    def __init__(self, spark, console, tbl):
        super().__init__(spark, console)
        self.tbl = tbl

    def update(self, selected=None):
        if selected is None:
            self._set_status("loading databases")
            result = [db.name for db in self.spark.catalog.listDatabases()]
            self.list.options = [""] + result
            self._clear_status()
        else:
            db = selected
            self.tbl.set_db(db)


class TableList(DBList):
    def __init__(self, spark, console, schema, data, sample_size=20):
        super().__init__(spark, console)
        self.schema = schema
        self.data = data
        self.sample_size = sample_size

    def _display_schema(self, schema, sep=False):
        if not sep:
            self.schema.clear_output()
        with self.schema:
            if sep:
                display(HTML("<hr>"))
            display(schema)

    def _display_data(self, data):
        self.data.clear_output()
        with self.data:
            display(data)

    def set_db(self, db):
        self.db = db
        self.update()

    def to_dict(self, c):
        return {
            "name": c.name,
            "data_type": c.dataType,
            "description": c.description,
            "nullable": c.nullable,
            "is_partition": c.isPartition,
            "is_bucket": c.isBucket,
        }

    def update(self, selected=None):
        tbl = selected
        if self.db is None:
            result = []
        elif tbl is None:
            self._set_status("loading tables")
            result = [tbl.name for tbl in self.spark.catalog.listTables(self.db)]
            self.list.options = [""] + result
            self._clear_status()
        else:
            # show schema
            self.schema.clear_output()
            self.data.clear_output()
            df = None
            try:
                df = pd.DataFrame(
                    [self.to_dict(col) for col in self.spark.catalog.listColumns(tbl, self.db)]
                )
            except:  # pylint: disable=bare-except
                self._display_schema("Cannot load schema")

            if df is None or df.empty:
                self._display_schema("Cannot load schema")
            else:
                df = df[
                    ["name", "data_type", "description", "nullable", "is_partition", "is_bucket"]
                ]
                self._display_schema(df)

                # show extended information
                pdf = self.spark.sql("describe extended %s.%s" % (self.db, tbl)).toPandas()
                pdf = pdf[pdf.index > len(df) + 1][["col_name", "data_type"]]
                pdf.columns = ["Key", "Value"]
                self._display_schema(pdf.style.hide_index(), sep=True)

                # show sample data
                self._set_status("loading data")
                try:
                    result = self.spark.sql(
                        "select * from %s.%s limit %d" % (self.db, tbl, self.sample_size)
                    ).toPandas()
                    self._display_data(result)
                except:  # pylint: disable=bare-except
                    self._display_data("Cannot load data")
                self._clear_status()


class Databases:
    """Database browser implementation

    Args:
        spark (SparkSession): Spark Session object
    """

    def __init__(self, spark):
        self.spark = spark

    def create(self, height=300, db_width=150, tbl_width=200, schema_width=700):
        console = Output(layout={"height": "28px"})
        schema = Output(
            layout={
                "height": _px(height),
                "width": _px(schema_width),
                "border": "1px solid grey",
                "overflow": "scroll",
            }
        )
        data = Output()

        tbl = TableList(self.spark, console, schema, data)
        tbl.create(width=tbl_width, height=height)

        db = DatabaseList(self.spark, console, tbl)
        db.create(width=db_width, height=height)

        display(
            VBox(
                [
                    HBox(
                        [
                            VBox([db.refresh, db.list]),
                            VBox([tbl.refresh, tbl.list]),
                            VBox([console, schema]),
                        ]
                    ),
                    data,
                ]
            )
        )



import os
import threading
import time

from IPython.display import display
from IPython import get_ipython

try:
    from ipywidgets import FloatProgress, Label, HBox, VBox, Layout, Button

    has_ipywidgets = True
except ImportError:
    has_ipywidgets = False


def debug():
    return os.environ.get("PROGRESS_DEBUG", None) == "TRUE"


class ProgressBars:
    def __init__(self):
        self.style = {"description_width": "initial"}
        self.bars = {}
        self.labels = {}
        self.last_job = None

        self.progressbars = VBox([])

        cancel_button = Button(button_style="", tooltip="Cancel Spark Job", icon="window-close")
        cancel_button.add_class("db-button")

        toggle_button = Button(
            button_style="", tooltip="Toggle progress bar", icon="arrow-circle-right"
        )
        toggle_button.add_class("db-button")
        toggle_button.on_click(self._toggle(self.progressbars))

        self.indicator = HBox([toggle_button, self.progressbars])
        self.progressbar_showing = False

    def _toggle(self, widget):
        def f(b):
            for w in widget.children:
                h = w.layout.height
                if h is None or h == "16px":
                    w.layout.height = "0px"
                    b.icon = "arrow-circle-down"
                else:
                    w.layout.height = "16px"
                    b.icon = "arrow-circle-right"

        return f

    def bar_needed(self, job):
        return self.bars.get(job, None) is None

    def add_bar(self, job):
        self.bars[job] = FloatProgress(
            value=0.0,
            min=0.0,
            max=100.0,
            description="Job: %04d Stage: %04d" % (job, 0),
            bar_style="info",
            orientation="horizontal",
            style=self.style,
        )
        self.bars[job].add_class("db-bar")
        self.labels[job] = Label(
            value="",
            description="Code:",
            disabled=False,
            layout=Layout(width="800px", height="100%", margin="0 0 0 5px"),
        )
        self.labels[job].add_class("db-label")

        progressbar = HBox([self.bars[job], self.labels[job]])
        self.progressbars.children = self.progressbars.children + (progressbar,)
        if not self.progressbar_showing:
            self.progressbar_showing = True
            display(self.indicator)

    def set_last_job(self, job):
        self.last_job = job

    def finish_last_job(self):
        if self.last_job is not None:
            self.bars[self.last_job].value = 100.0

    def update_bar(self, job, description, label):
        self.bars[job].description = description
        self.labels[job].value = label

    def update_progress(self, job, value):
        self.bars[job].value = value


class Progress:
    def __init__(self, sc, job_info):
        self.sc = sc
        self.job_info = job_info
        self.thread = None
        self.tracker = sc.statusTracker()
        self.progressbars = ProgressBars()

    def start(self):
        self.job_info.dump("START")
        self.thread = threading.Thread(target=self.worker, args=(self.progressbars,), daemon=True)
        self.job_info.current_thread = self.thread
        self.thread.start()

        # Re-attach to ensure main thread has group_id
        # Sometimes this was reset to last group_id, reason unclear
        self.job_info.attach()
        self.job_info.dump("POST START")

    def worker(self, progressbars):
        self.job_info.attach()
        self.job_info.is_running[self.job_info.group_id] = True
        self.job_info.dump("RUNNING")

        while self.job_info.is_running[self.job_info.group_id]:
            jobs = [
                (jobid, self.tracker.getJobInfo(jobid))
                for jobid in self.tracker.getJobIdsForGroup(self.job_info.group_id)
                if self.tracker.getJobInfo(jobid).status == "RUNNING"
            ]
            if debug():
                print("==> jobs", jobs, self.job_info.group_id)
            for j, job in jobs:
                if progressbars.bar_needed(j):
                    progressbars.finish_last_job()
                    progressbars.add_bar(j)
                    progressbars.set_last_job(j)

                stageIds = sorted(job.stageIds)
                for s in stageIds:
                    stageInfo = self.tracker.getStageInfo(s)
                    description = "Job: %04d Stage: %04d" % (j, s)
                    label = "code: '%s' / stages: %s" % (stageInfo.name, str(stageIds)[1:-1])
                    progressbars.update_bar(j, description, label)
                    if stageInfo.numActiveTasks > 0:
                        progress = int(100 * stageInfo.numCompletedTasks / stageInfo.numTasks)
                        progressbars.update_progress(j, progress)
            time.sleep(0.2)

        progressbars.finish_last_job()

        self.job_info.dump("LEFT LOOP")


def is_running(tracker, group_id):
    job_status = [tracker.getJobInfo(jobid).status for jobid in tracker.getJobIdsForGroup(group_id)]
    return any(status == "RUNNING" for status in job_status)


def start_wrapper(sc, job_info):
    def start_progressbar(_):
        job_info.new_group_id()
        job_info.dump("INIT")
        prog_bar = Progress(sc, job_info)
        prog_bar.start()

    return start_progressbar


def stop_wrapper(sc, job_info):
    def stop_progressbar(_):
        job_info.is_running[job_info.group_id] = False

        # job_info.attach()
        job_info.dump("STOPPED")

        tracker = sc.statusTracker()

        if is_running(tracker, job_info.group_id):
            print("Killing spark job ", end="")
            while is_running(tracker, job_info.group_id):
                sc.cancelJobGroup(job_info.group_id)
                print(".", end="")
                time.sleep(1)
            print(". Spark job killed")

        job_info.stop_all()
        if job_info.current_thread is not None:
            if debug():
                print("\nWAITING for thread to stop")
            job_info.current_thread.join()
            if debug():
                print("\nTHREAD STOPPED")

    return stop_progressbar


def load_progressbar(sc, job_info):
    def unregister(ip, event, func_name):
        f = [f for f in ip.events.callbacks[event] if f.__name__ == func_name]
        if len(f) > 0:
            ip.events.unregister(event, f[0])

    if has_ipywidgets:
        ip = get_ipython()
        for event, name, func in (
            ("pre_run_cell", "start_progressbar", start_wrapper(sc, job_info)),
            ("post_run_cell", "stop_progressbar", stop_wrapper(sc, job_info)),
        ):
            unregister(ip, event, name)
            ip.events.register(event, func)



import atexit
import datetime
import glob
import io
import os
import random
import sys
import threading
import uuid
import warnings

from IPython import get_ipython
from IPython.display import HTML, display
from IPython.core.magic import Magics, magics_class, line_cell_magic

from databrickslabs_jupyterlab.rest import Command
from databrickslabs_jupyterlab.progress import load_progressbar, debug
from databrickslabs_jupyterlab.dbfs import Dbfs
from databrickslabs_jupyterlab.database import Databases
from databrickslabs_jupyterlab.notebook import Notebook

py4j = glob.glob("/databricks/spark/python/lib/py4j-*-src.zip")[0]
sys.path.insert(0, py4j)
sys.path.insert(0, "/databricks/spark/python")
sys.path.insert(0, "/databricks/jars/spark--driver--spark--resources-resources.jar")

from pyspark.conf import SparkConf  # pylint: disable=import-error,wrong-import-position

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    # Suppress py4j loading message on stderr by redirecting sys.stderr
    stderr_orig = sys.stderr
    sys.stderr = io.StringIO()
    from PythonShell import get_existing_gateway, RemoteContext  # pylint: disable=import-error

    out = sys.stderr.getvalue()
    # Restore sys.stderr
    sys.stderr = stderr_orig
    # Print any other error message to stderr
    if not "py4j imported" in out:
        print(out, file=sys.stderr)

from dbutils import DBUtils  # pylint: disable=import-error,wrong-import-position


class JobInfo:
    """Job info class for Spark jobs
    Args:
        sc (SparkContext): Spark Context
    """

    def __init__(self, sc):
        self.pool_id = str(random.getrandbits(64))
        self.group_id = "jupyterlab-default-group"
        self.is_running = {}
        self.current_thread = None
        self.sc = sc

    def dump(self, tag):
        if debug():
            print(
                tag,
                "%s (%s)"
                % (
                    datetime.datetime.now().isoformat(),
                    threading.current_thread().__class__.__name__,
                ),
            )
            print("%s: pool_id              %s" % (tag, self.pool_id))
            print("%s: group_id             %s" % (tag, self.group_id))
            print(
                "%s: spark.scheduler.pool %s"
                % (tag, self.sc.getLocalProperty("spark.scheduler.pool"))
            )
            print(
                "%s: spark.jobGroup.id    %s" % (tag, self.sc.getLocalProperty("spark.jobGroup.id"))
            )
            print("%s: running              %s" % (tag, self.is_running.get(self.group_id, None)))

    def attach(self):
        if debug():
            print(
                "\nATTACHING to ",
                self.group_id,
                "in",
                threading.current_thread().__class__.__name__,
            )
        # Catch [SPARK-22340][PYTHON] Add a mode to pin Python thread into JVM's
        # This code explicitely attaches both python threads to the value Spark Context is set to
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.sc.setLocalProperty("spark.scheduler.pool", self.pool_id)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.sc.setJobGroup(self.group_id, "jupyterlab job group", True)

    def stop_all(self):
        if debug():
            print("\nSTOPPING all running threads")
        for k, v in self.is_running.items():
            if v and debug():
                print("- Job to be stopped:", k)
            self.is_running[k] = False

    def new_group_id(self):
        self.group_id = self.pool_id + "_" + uuid.uuid4().hex
        self.attach()


class DbjlUtils:
    def __init__(self, shell, entry_point):
        self._dbutils = DBUtils(shell, entry_point)
        self.fs = self._dbutils.fs
        self.secrets = self._dbutils.secrets
        self.notebook = Notebook()

    def help(self):
        html = """
        This module provides a subset of the DBUtils tools working for Jupyterlab Integration
        <br/><br/>
        <b>fs: DbfsUtils</b> -&gt; Manipulates the Databricks filesystem (DBFS) from the console
        <br/>
        <b>secrets: SecretUtils</b> -&gt; Provides utilities for leveraging secrets within notebooks
        <br/>
        <b>notebook: NotebookUtils</b> -&gt; Utilities for the control flow of a notebook (EXPERIMENTAL)
        """
        display(HTML(html))


@magics_class
class DbjlMagics(Magics):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = os.environ["DBJL_HOST"]
        self.cluster = os.environ["DBJL_CLUSTER"]
        self.scope = None
        self.key = None
        self.command = None

    @line_cell_magic
    def scala(self, line, cell=None):
        "Magic that allows to execute scala in a notebook"
        if cell is None:
            print("Use %%scala as a cell magic")
        else:
            try:
                command = get_ipython().user_ns["__scalaCommand"]
                result = command.execute(cell, full_result=True)
            except Exception as ex:  # pylint: disable=broad-except
                result = (-1, "%s: %s" % (type(ex), ex))

            if result[0] == 0:
                typ = result[1]["results"]["resultType"]
                data = result[1]["results"]["data"]
                if typ == "text":
                    print(data)
                elif typ == "table":
                    import pandas as pd

                    df = pd.DataFrame(data)
                    df.columns = [c["name"] for c in result[1]["results"]["schema"]]
                    display(df)
                    if result[1]["results"]["truncated"]:
                        print("Data truncated, more than 1000 rows")
            else:
                print("Error: " + result[1]["results"]["cause"])

    @line_cell_magic
    def sql(self, line, cell=None):
        """Cell magic th execute SQL commands

        Args:
            line (str): line behind %sql
            cell (str, optional): cell below %sql. Defaults to None.
        
        Returns:
            DataFrame: DataFrame of the SQL result
        """
        ip = get_ipython()
        spark = ip.user_ns["spark"]
        if cell is None:
            code = line
        else:
            code = cell
        if "explain" in code.lower():
            result = spark.sql(code).collect()
            for plan in result:
                print(plan.plan)
        else:
            return spark.sql(code)

    @line_cell_magic
    def fs(self, line, cell=None):
        """Cell magic to browse DBFS

        Args:
            line (str): line behind %fs
            cell (str, optional): cell below %%fs. Defaults to None.
        
        Returns:
            None
        """

        def confirm(prompt):
            answer = ""
            while answer not in ["y", "n"]:
                answer = input(prompt + " (y/n) ").lower()
            return answer == "y"

        ip = get_ipython()
        dbutils = ip.user_ns["dbutils"]

        if cell is None:
            cmd = line.strip().split()
        else:
            cmd = cell.strip().split()

        if cmd[0] == "cp":
            if cmd[1] in ("--recurse", "-r"):
                if confirm("Recursively copy '{}' to '{}'".format(cmd[2], cmd[3])):
                    dbutils.fs.cp(cmd[2], cmd[3], recurse=True)
                    print("done")
            else:
                if confirm("Copy '{}' to '{}'".format(cmd[1], cmd[2])):
                    dbutils.fs.cp(cmd[1], cmd[2], recurse=False)
                    print("done")

        elif cmd[0] == "head":
            if cmd[1] in ("--maxBytes", "-m"):
                print("Showing head of file '{}' with max. {} bytes:\n".format(cmd[3], cmd[2]))
                print(dbutils.fs.head(cmd[3], max_bytes=int(cmd[2])))
            else:
                print("Show head of file '{}':\n".format(cmd[1]))
                print(dbutils.fs.head(cmd[1]))

        elif cmd[0] == "ls":
            print("List directory '{}':\n".format(cmd[1]))
            for f in dbutils.fs.ls(cmd[1]):
                print(f)

        elif cmd[0] == "mkdirs":
            if confirm("Make dirs '{}'".format(cmd[1])):
                dbutils.fs.mkdirs(cmd[1])
                print("done")

        elif cmd[0] == "mv":
            if cmd[1] in ("--recurse", "-r"):
                if confirm("Recursively moving '{}' to '{}'".format(cmd[2], cmd[3])):
                    dbutils.fs.mv(cmd[2], cmd[3], recurse=True)
                    print("done")
            else:
                if confirm("Moving '{}' to '{}'".format(cmd[1], cmd[2])):
                    dbutils.fs.mv(cmd[1], cmd[2], recurse=False)
                    print("done")

        elif cmd[0] == "put":
            if cmd[1] in ("--overwrite", "-o"):
                if confirm("Overwriting file '{}' with '{}'".format(cmd[2], " ".join(cmd[3:]))):
                    dbutils.fs.put(cmd[2], " ".join(cmd[3:]), overwrite=True)
                    print("done")
            else:
                if confirm("Writing '{}' to file '{}'".format(" ".join(cmd[2:]), cmd[1])):
                    dbutils.fs.put(cmd[1], " ".join(cmd[2:]), overwrite=False)
                    print("done")

        elif cmd[0] == "rm":
            if cmd[1] in ("--recurse", "-r"):
                if confirm("Recursively deleting '{}'".format(cmd[2])):
                    dbutils.fs.rm(cmd[2], recurse=True)
                    print("done")
            else:
                if confirm("Deleting '{}'".format(cmd[1])):
                    dbutils.fs.rm(cmd[1], recurse=False)
                    print("done")


class DatabricksBrowser:
    """[summary]
    Args:
        spark (SparkSession): Spark Session object
        dbutils (DBUtils): DbUtils object
    """

    def __init__(self, spark, dbutils):
        self.spark = spark
        self.dbutils = dbutils

    def dbfs(self, path="/", height="400px"):
        """Start dbfs browser"""
        Dbfs(self.dbutils).create(path, height)

    def databases(self):
        """Start Database browser"""
        Databases(self.spark).create()

    def experiments(self, experiment_name):
        """Start the experiment browser for a given experiment name"""
        from databrickslabs_jupyterlab.mlflow import MLflowBrowser

        return MLflowBrowser(experiment_name)


def dbcontext(progressbar=True, gw_port=None, gw_token=None, token=None, scala_context_id=None):
    """Create a databricks context
    The following objects will be created
    - Spark Session
    - Spark Context
    - Spark Hive Context
    - DBUtils (fs module only)
    
    Args:
        progressbar (bool, optional): If True spark progressbars will be shown. Default: True.
    """

    def get_sparkui_url(host, organisation, clusterId):
        if organisation is None:
            sparkUi = "%s#/setting/clusters/%s/sparkUi" % (host, clusterId)
        else:
            sparkUi = "%s/?o=%s#/setting/clusters/%s/sparkUi" % (host, organisation, clusterId)
        return sparkUi

    def show_status(spark):
        output = """
        <div>
            <dl>
            <dt>Spark Version</dt><dd>{sc.version}</dd>
            <dt>Spark Application</dt><dd>{sc.appName}</dd>
            <dt>Spark UI</dt><dd><a href="{sparkUi}">go to ...</a></dd>
            </dl>
        </div>
        """.format(
            sc=spark.sparkContext,
            sparkUi=get_sparkui_url(host, organisation, clusterId),
            # num_executors=len(spark.sparkContext._jsc.sc().statusTracker().getExecutorInfos()),
        )
        display(HTML(output))

    # Get the configuration injected by the client
    #
    profile = os.environ.get("DBJL_PROFILE", None)
    host = os.environ.get("DBJL_HOST", None)
    clusterId = os.environ.get("DBJL_CLUSTER", None)
    organisation = os.environ.get("DBJL_ORG", None)

    print(
        "Remote init: profile={}, organisation={}. cluster_id={}, host={}".format(
            profile, organisation, clusterId, host
        )
    )
    sparkUi = get_sparkui_url(host, organisation, clusterId)

    print("Remote init: Spark UI = {}".format(sparkUi))

    ip = get_ipython()

    if os.environ.get("DEBUG") == "DEBUG":
        print("Remote init: Gateway port = {}, token = {}".format(gw_port, gw_token))

    interpreter = "/databricks/python/bin/python"
    # Ensure that driver and executors use the same python
    #
    os.environ["PYSPARK_PYTHON"] = interpreter
    os.environ["PYSPARK_DRIVER_PYTHON"] = interpreter

    # ... and connect to this gateway
    #
    try:
        # up to DBR version 6.4
        gateway = get_existing_gateway(gw_port, True, gw_token)
    except TypeError:
        # for DBR 6.5 and higher
        gateway = get_existing_gateway(gw_port, True, gw_token, False)

    print("Remote init: Connected")

    # Retrieve spark session, sqlContext and sparkContext
    #

    conf = SparkConf(_jconf=gateway.entry_point.getSparkConf())
    sc = RemoteContext(gateway=gateway, conf=conf)
    if sc.version < "3.0":
        from pyspark.sql import HiveContext  # pylint: disable=import-error,import-outside-toplevel

        sqlContext = HiveContext(sc, gateway.entry_point.getSQLContext())
    else:
        from pyspark.sql import SQLContext  # pylint: disable=import-error,import-outside-toplevel
        from pyspark.sql.session import (  # pylint: disable=import-error,import-outside-toplevel
            SparkSession,
        )

        jsqlContext = gateway.entry_point.getSQLContext()
        sqlContext = SQLContext(sc, SparkSession(sc, jsqlContext.sparkSession()), jsqlContext)

    spark = sqlContext.sparkSession
    sc = spark.sparkContext

    print("Remote init: Spark Session created")

    # Enable pretty printing of dataframes
    #
    spark.conf.set("spark.sql.repl.eagerEval.enabled", "true")

    # Define a separate pool for the fair scheduler
    #
    job_info = JobInfo(sc)

    # Patch the remote spark UI into the _repr_html_ call
    #
    def repr_html(uiWebUrl):
        def sc_repr_html():
            return """
            <div>
                <p><b>SparkContext</b></p>
                <p><a href="{uiWebUrl}">Spark UI</a></p>
                <dl>
                  <dt>Version</dt><dd><code>v{sc.version}</code></dd>
                  <dt>AppName</dt><dd><code>{sc.appName}</code></dd>
                  <dt>Master</dt><dd><code>{sc.master}</code></dd>
                </dl>
            </div>
            """.format(
                sc=spark.sparkContext, uiWebUrl=uiWebUrl
            )

        return sc_repr_html

    sc_repr_html = repr_html(sparkUi)
    sc._repr_html_ = sc_repr_html  # pylint: disable=protected-access

    # Monkey patch Databricks Cli to allow mlflow tracking with the credentials provided
    # by this routine
    # Only necessary when mlflow is installed
    #

    print("Remote init: Configuring mlflow")
    try:
        from databricks_cli.configure.provider import (  # pylint: disable=import-outside-toplevel
            ProfileConfigProvider,
            DatabricksConfig,
        )

        def get_config(self):  # pylint: disable=unused-argument
            config = DatabricksConfig(host, None, None, token, False)
            if config.is_valid:
                return config
            return None

        ProfileConfigProvider.get_config = get_config
    except Exception:  # pylint: disable=broad-except
        print("Cannot initialize mlflow")

    # Initialize the ipython shell with spark context
    #
    shell = get_ipython()
    shell.sc = sc
    shell.sqlContext = sqlContext
    shell.displayHTML = lambda html: display(HTML(html))
    shell.user_ns["displayHTML"] = lambda html: display(HTML(html))

    # Retrieve the py4j gateway entrypoint
    #
    entry_point = spark.sparkContext._gateway.entry_point  # pylint: disable=protected-access

    # Initialize dbutils
    #
    dbutils = DbjlUtils(shell, entry_point)

    # Setting up Spark progress bar
    #
    if progressbar:
        # print("Set up Spark progress bar")
        load_progressbar(sc, job_info)

    # Register sql magic
    #
    # ip.register_magic_function(sql, magic_kind="line_cell")
    ip.register_magics(DbjlMagics)

    # Setup scala context

    if scala_context_id is not None:
        print("Remote init: Configuring scala Command")
        try:
            scalaCommand = Command(
                url=host,
                cluster_id=clusterId,
                token=token,
                language="scala",
                scala_context_id=scala_context_id,
            )
            ip.user_ns["__scalaCommand"] = scalaCommand
        except:  # pylint: disable=bare-except
            print("Cannot create scala command, so %%scala will not work")
            scalaCommand = None
    else:
        print("Remote Init: %%scala will not work")

    # Forward spark variables to the user namespace
    #
    ip.user_ns["spark"] = spark
    ip.user_ns["sc"] = sc
    ip.user_ns["sqlContext"] = sqlContext
    ip.user_ns["dbutils"] = dbutils
    ip.user_ns["dbbrowser"] = DatabricksBrowser(spark, dbutils)

    print("Remote init: The following global variables have been created:")
    print("- spark       Spark session")
    print("- sc          Spark context")
    print("- sqlContext  Hive Context")
    print("- dbutils     Databricks utilities (filesystem access only)")
    print("- dbbrowser   Allows to browse dbfs and databases:")
    print("              - dbbrowser.dbfs()")
    print("              - dbbrowser.databases()\n")

    show_status(spark)
    return None



from dataclasses import dataclass
from datetime import datetime
from difflib import ndiff
from typing import List, Dict
import os
import re


@dataclass
class Comment:
    indent: str
    value: str

    def to_string(self):
        return f"{self.indent}{self.value}"


@dataclass
class KeyValue:
    indent: str
    key: str
    sep: str
    value: str
    comment: str = ""

    def to_string(self):
        return f"{self.indent}{self.key}{self.sep}{self.value}{self.comment}"


@dataclass
class Block:
    head: KeyValue
    params: List[KeyValue]
    indent: str = ""

    def to_string(self):
        return f"{self.head.to_string()}\n" + "\n".join([p.to_string() for p in self.params])

    def get_param(self, key):
        for kv in self.params:
            if isinstance(kv, KeyValue) and kv.key.lower() == key.lower():
                return kv
        return None

    def set_param(self, key, value):
        p = self.get_param(key)
        if p is None:
            if self.params and self.params[-1].value == "":
                self.params.insert(
                    len(self.params) - 1,
                    KeyValue(self.indent, key, " ", value),
                )
            else:
                self.params.append(KeyValue(self.indent, key, " ", value))
        else:
            p.value = value

        return self


@dataclass
class SshConfig:
    config_file: str
    blocks: List[Block] = None

    def __post_init__(self):
        self.load(self.config_file)
        self.parse()

    def to_string(self):
        self.config_file = os.path.expanduser(self.config_file)
        return "\n".join([b.to_string() for b in self.blocks])

    def load(self, config):
        self.data = ""
        if os.path.exists(self.config_file):
            with open(self.config_file, "r") as fd:
                self.data = fd.read()

    def write(self):
        with open(self.config_file, "w") as fd:
            fd.write(self.to_string())

    @staticmethod
    def parse_line(raw_line):
        indent, line = re.match(r"(^\s*)(.*)", raw_line).groups()
        definition = re.split(r"(\s*#[#\s]+)", line)
        comment = "" if len(definition) == 1 else "".join(definition[1:])
        key_values = definition[0]

        values = re.split(r"([=\s]+)", key_values)
        key = values[0]
        sep = "" if len(values) == 1 else values[1]
        value = "" if len(values) == 1 else "".join(values[2:])
        return [indent, key, sep, value, comment]

    def parse(self):
        self.blocks = []
        lines = self.data.split("\n")
        block = None
        for line in lines:
            indent, key, sep, params, comment = self.parse_line(line)
            if key == "" and comment == "":
                if block is None:
                    self.blocks.append(Comment("", ""))
                else:
                    block.params.append(Comment("", ""))
            elif key == "" and comment != "" and block is None:
                self.blocks.append(Comment(indent, comment))
            elif key.lower() in ["host", "match"]:
                if block is not None:
                    self.blocks.append(block)
                    block = None
                head = KeyValue(indent, key, sep, params, comment)
                block = Block(head, [])
            else:
                block.params.append(KeyValue(indent, key, sep, params, comment))
                if key != "":
                    block.indent = indent

        if block is not None:
            self.blocks.append(block)
        return self

    def get_host(self, name):
        for b in self.blocks:
            if isinstance(b, Block) and "".join(b.head.value) == name:
                return b
        return None

    def get_hosts(self):
        return self.blocks

    def add_host(self, host):
        block = self.get_host(host)
        if block is None:
            block = Block(
                KeyValue("", "Host", " ", host, ""),
                [],
                "    ",
            )
            self.blocks.append(block)
        return block

    def remove_host(self, name):
        self.blocks = [b for b in self.blocks if isinstance(b, Block) and b.head.value != name]

    def dump(self):
        print("\n".join(ndiff(self.data.split("\n"), self.to_string().split("\n"))))


import base64
import json
import os
import re
import time
from os.path import expanduser
import socket
from urllib.parse import unquote

import requests

from databricks_cli.configure.provider import ProfileConfigProvider
from databricks_cli.sdk.api_client import ApiClient
from databricks_cli.clusters.api import ClusterApi
from databricks_cli.libraries.api import LibrariesApi

# import databrickslabs_jupyterlab
from databrickslabs_jupyterlab._version import __version__
from databrickslabs_jupyterlab.rest import Command, DatabricksApiException
from databrickslabs_jupyterlab.utils import (
    bye,
    print_ok,
    print_error,
    print_warning,
    question,
    utf8_decode,
    execute,
)
from databrickslabs_jupyterlab.local import execute, get_local_libs


def connect(profile):
    """Initialize Databricks API client

    Args:
        profile (str): Databricks CLI profile string

    Returns:
        ApiClient: Databricks ApiClient object
    """
    config = ProfileConfigProvider(profile).get_config()
    if config is None:
        print_error("Cannot initialize ApiClient")
        bye(1)
    verify = config.insecure is None
    if config.is_valid_with_token:
        api_client = ApiClient(host=config.host, token=config.token, verify=verify)
        api_client.default_headers["user-agent"] = "databrickslabs-jupyterlab-%s" % __version__
        return api_client
    else:
        print_error(
            "No token found for profile '%s'.\nUsername/password in .databrickscfg is not supported by databrickslabs-jupyterlab"
            % profile
        )
        bye(1)


def select_cluster(clusters):
    """Build a list of clusters for the user to select one

    Args:
        clusters (list): list of clusters

    Returns:
        dict: Cluster config of selected cluster
    """

    def entry(i, cluster):
        if cluster.get("autoscale", None) is None:
            return "%s: '%s' (id: %s, state: %s, workers: %d)" % (
                i,
                cluster["cluster_name"],
                cluster["cluster_id"],
                cluster["state"],
                cluster["num_workers"],
            )
        else:
            return "%s: '%s' (id: %s, state: %s, scale: %d-%d)" % (
                i,
                cluster["cluster_name"],
                cluster["cluster_id"],
                cluster["state"],
                cluster["autoscale"]["min_workers"],
                cluster["autoscale"]["max_workers"],
            )

    answer = question(
        "cluster_id",
        "Which cluster to connect to?",
        [entry(i, cluster) for i, cluster in enumerate(clusters)],
    )
    if answer["cluster_id"] is None:
        return None
    else:
        return clusters[int(answer["cluster_id"].split(":")[0])]


def get_cluster(profile, cluster_id=None, status=None):
    """Get the cluster configuration from remote

    Args:
        profile (str): Databricks CLI profile string
        cluster_id (str, optional): If cluster_id is given, users will not be asked to select one.
                                    Defaults to None.
        status (Status, optional): A Status class providing set_status. Defaults to None.
        tunnel (str, optional): An ssh tunne geiven as "address:port"

    Returns:
        tuple: Cluster configs: cluster_id, public_ip, cluster_name and a flag
                                whether it was started

    Returns:
        tuple: (cluster_id, public_ip, cluster_name, started)
    """

    if os.environ.get("SSH_TUNNEL") is None:
        tunnel = None
    else:
        tunnel = os.environ.get("SSH_TUNNEL")
        print_warning("\n  ==> Using tunnel: " + tunnel + "\n")

    failure = (None, None, None, None)

    with open("%s/.ssh/id_%s.pub" % (expanduser("~"), profile)) as fd:
        try:
            ssh_pub = fd.read().strip()
        except:  # pylint: disable=bare-except
            print_error(
                "   Error: ssh key for profile 'id_%s.pub' does not exist in %s/.ssh"
                % (profile, expanduser("~"))
            )
            bye()

    try:
        apiclient = connect(profile)
        client = ClusterApi(apiclient)
        clusters = client.list_clusters()
    except Exception as ex:  # pylint: disable=broad-except
        print_error(ex)
        return failure

    clusters = clusters["clusters"]

    if cluster_id is not None:
        cluster = None
        for c in clusters:
            if c["cluster_id"] == cluster_id:
                cluster = c
                break

        if cluster is None:
            print_error(
                "   Error: A cluster with id '%s' does not exist in the workspace of profile '%s'"
                % (cluster_id, profile)
            )
            return failure

        if ssh_pub not in [c.strip() for c in cluster.get("ssh_public_keys", [])]:
            print_error(
                "   Error: Cluster with id '%s' does not have ssh key '~/.ssh/id_%s' configured"
                % (cluster_id, profile)
            )
            return failure
    else:
        my_clusters = [
            cluster
            for cluster in clusters
            if ssh_pub in [c.strip() for c in cluster.get("ssh_public_keys", [])]
        ]

        if not my_clusters:
            print_error(
                (
                    "    Error: There is no cluster in the workspace for profile '%s' "
                    + "configured with ssh key '~/.ssh/id_%s':"
                )
                % (profile, profile)
            )
            print(
                (
                    "    Use 'databrickslabs_jupyterlab %s -s' to configure ssh for clusters "
                    + "in this workspace\n"
                )
                % profile
            )
            return failure

        current_conda_env = os.environ.get("CONDA_DEFAULT_ENV", None)
        found = None
        ind = -1
        for i, c in enumerate(my_clusters):
            if c["cluster_name"].replace(" ", "_") == current_conda_env:
                found = c["cluster_name"]
                ind = i
                break

        if found is not None:
            print_warning(
                ("\n   => The current conda environment is '%s'. " + current_conda_env)
                + (
                    "\n      You might want to select cluster %d with the name '%s'?\n"
                    % (ind, found)
                )
            )

        cluster = select_cluster(my_clusters)
        if cluster is None:
            return failure

    cluster_id = cluster["cluster_id"]
    cluster_name = cluster["cluster_name"]

    try:
        response = client.get_cluster(cluster_id)
    except Exception as ex:  # pylint: disable=broad-except
        print_error(ex)
        return failure

    state = response["state"]
    if not state in ["RUNNING", "RESIZING"]:
        if state == "TERMINATED":
            print("   => Starting cluster %s" % cluster_id)
            if status is not None:
                status.set_status(profile, cluster_id, "Cluster Starting")
            try:
                response = client.start_cluster(cluster_id)
            except Exception as ex:  # pylint: disable=broad-except
                print_error(ex)
                return failure

        print("   => Waiting for cluster %s being started (this can take up to 5 min)" % cluster_id)
        print("   ", end="", flush=True)

        while not state in ("RUNNING", "RESIZING"):
            if status is not None:
                status.set_status(profile, cluster_id, "Cluster Starting", False)
            print(".", end="", flush=True)
            time.sleep(5)
            try:
                response = client.get_cluster(cluster_id)
            except Exception as ex:  # pylint: disable=broad-except
                print_error(ex)
                return failure

            if response.get("error", None) is not None:
                print_error(response["error"])
                return failure
            else:
                state = response["state"]
        print_ok("\n   => OK")

        if status is not None:
            status.set_status(profile, cluster_id, "Cluster started")

        print(
            "\n   => Waiting for libraries on cluster %s being installed (this can take some time)"
            % cluster_id
        )
        print("   ", end="", flush=True)

        done = False
        while not done:
            try:
                states = get_library_state(profile, cluster_id)
            except DatabricksApiException as ex:
                print_error(ex)
                return failure

            installing = any([s in ["PENDING", "RESOLVING", "INSTALLING"] for s in states])
            if installing:
                if status is not None:
                    status.set_status(profile, cluster_id, "Installing cluster libraries", False)
                print(".", end="", flush=True)
                time.sleep(5)
            else:
                done = True
                print_ok("\n   => OK\n")

        if status is not None:
            status.set_status(profile, cluster_id, "Cluster libraries installed", False)

    if tunnel is not None:
        address = tunnel
    else:
        public_dns = response["driver"].get("public_dns")
        if public_dns is None:
            print_error("   Error: Cluster does not have public DNS name")
            return failure
        address = "%s:%s" % (public_dns, 2200)

    print_ok("   => Selected cluster: %s (%s)" % (cluster_name, address))

    return (cluster_id, address, cluster_name, None)


def get_python_path(host, conda_env=None):  # pylint: disable=unused-argument
    python_path = "/databricks/python/bin"
    return python_path


def install_libs(cluster_id, host, token):
    """Install ipywidgets, databrickslabs_jupyterlab libraries on the driver

    Args:
        host (str): host from databricks cli config for given profile string
        python_path (str): Remote python path to be used for kernel
    """
    python_path = get_python_path(cluster_id)

    packages = get_local_libs()
    deps = {p["name"]: p["version"] for p in packages if p["name"] in ["ipywidgets", "ipykernel"]}
    libs = [
        "ipywidgets==%s" % deps["ipywidgets"],
        "ipykernel==%s" % deps["ipykernel"],
        "databrickslabs-jupyterlab==%s" % __version__,
        "pygments>=2.4.1",
    ]

    print("   => Installing ", " ".join(libs))

    ssh = os.environ.get("SSH") or "ssh"
    cmd = [ssh, cluster_id, "sudo", python_path + "/pip", "install", "--upgrade"] + libs
    result = execute(cmd)

    if result["returncode"] == 0:
        if result["stdout"].startswith("Requirement already satisfied"):
            return (0, "Requirement already satisfied")
        else:
            return (0, "Installed")

    return (result["returncode"], result["stdout"] + "\n" + result["stderr"])


def get_remote_packages(cluster_id, host, token):
    cmd = (
        "import sys, pkg_resources, json; "
        + 'print(json.dumps([{"name": "python", "version": "%d.%d" % (sys.version_info.major, sys.version_info.minor)}] + '
        + '[{"name":p.key, "version":p.version} for p in pkg_resources.working_set]))'
    )
    try:
        command = Command(url=host, cluster_id=cluster_id, token=token)
        result = command.execute(cmd)
        command.close()
    except DatabricksApiException as ex:
        print(ex)
        return None
    return result


def is_reachable(endpoint):
    """Check whether a remote cluster is reachable

    Args:
        endpoint (str): Public IP address or DNS name and port as "address:port"

    Returns:
        bool: True if reachable else False
    """
    # return subprocess.call(["nc", "-z", address, str(port)]) == 0
    address, port = endpoint.split(":")
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(3)
    try:
        result = sock.connect_ex((address, int(port)))
        result = result == 0
    except:  # pylint: disable=bare-except
        result = False

    sock.close()
    return result


def get_library_state(profile, cluster_id):
    """Get the state of the library installation on the remote cluster

    Args:
        cluster_id (str): Cluster ID
        host (str): host from databricks cli config for given profile string
        token (str): token from databricks cli config for given profile stringf

    Returns:
        list: list of installation status for each custom library
    """
    try:
        apiclient = connect(profile)
        client = LibrariesApi(apiclient)
        libraries = client.cluster_status(cluster_id)
    except Exception as ex:  # pylint: disable=broad-except
        print_error(ex)
        return None

    if libraries.get("library_statuses", None) is None:
        return []
    else:
        return [lib["status"] for lib in libraries["library_statuses"]]


def check_installed(cluster_id, host, token):
    """Check whether databrickslabs_jupyterlab is installed on the remote host

    Args:

    Returns:
        bool: True if installed else False
    """
    result = get_remote_packages(cluster_id, host, token)
    if result[0] == 0:
        packages = json.loads(result[1])
        package_names = [p["name"] for p in packages]
        needed = ["databrickslabs-jupyterlab", "ipywidgets", "ipykernel"]

        return all([p in package_names for p in needed])
    else:
        print(result[1])
        return False


def version_check(cluster_id, host, token, flag):
    """Compare local and remote library versions

    Args:
        cluster_id (str): Cluster ID
        host (str): host from databricks cli config for given profile string
        token (str): token from databricks cli config for given profile string
        flag (str): all|diff|same
    """

    def normalize(key):
        return key.lower().replace("-", "_")

    packages = get_local_libs()
    deps = {normalize(p["name"]): p["version"] for p in packages}

    result = get_remote_packages(cluster_id, host, token)
    if result[0] == 0:
        remote_packages = json.loads(result[1])
    else:
        return
    remote_deps = {normalize(p["name"]): p["version"] for p in remote_packages}
    joint_keys = sorted(list(set(list(deps.keys()) + list(remote_deps.keys()))))
    print("%-30s %-10s%-10s" % ("Package", "local", "remote"))
    if str(flag) == "all":
        scope = joint_keys
    elif str(flag) == "same":
        scope = [key for key in joint_keys if deps.get(key, None) == remote_deps.get(key, None)]
    else:
        scope = [
            key
            for key in joint_keys
            if deps.get(key, None) != remote_deps.get(key, None)
            # and deps.get(key, None) is not None and remote_deps.get(key, None) is not None
        ]
    for key in scope:
        result = "%-30s %-10s  %-10s" % (key, deps.get(key, "--"), remote_deps.get(key, "--"))
        if deps.get(key) == remote_deps.get(key):
            print_ok(result)
        else:
            print_error(result)


def configure_ssh(profile, cluster_id):
    """Configure SSH for the remote cluster

    Args:
        profile (str): Databricks CLI profile string
        host (str): host from databricks cli config for given profile string
        token (str): token from databricks cli config for given profile string
        cluster_id (str): cluster ID
    """
    sshkey_file = os.path.expanduser("~/.ssh/id_%s" % profile)
    if not os.path.exists(sshkey_file):
        print("\n   => ssh key '%s' does not exist" % sshkey_file)
        answer = input("   => Shall it be created (y/n)? (default = n): ")
        if answer.lower() == "y":
            print("   => Creating ssh key %s" % sshkey_file)
            result = execute(["ssh-keygen", "-b", "2048", "-N", "", "-f", sshkey_file])
            if result["returncode"] == 0:
                print_ok("   => OK")
            else:
                print_error(result["stderr"])
                bye()
        else:
            bye()
    else:
        print_ok("\n   => ssh key '%s' already exists" % sshkey_file)

    with open(sshkey_file + ".pub", "r") as fd:
        sshkey = fd.read().strip()

    try:
        apiclient = connect(profile)
        client = ClusterApi(apiclient)
    except Exception as ex:  # pylint: disable=broad-except
        print_error(ex)
        return None

    try:
        response = client.get_cluster(cluster_id)
    except Exception as ex:  # pylint: disable=broad-except
        print_error(ex)
        return None

    ssh_public_keys = response.get("ssh_public_keys", [])
    if sshkey in [key.strip() for key in ssh_public_keys]:
        print_ok("   => public ssh key already configured for cluster %s" % cluster_id)
        bye()

    request = {}

    for key, val in response.items():
        if key not in [
            "driver",
            "executors",
            "spark_context_id",
            "state",
            "state_message",
            "start_time",
            "terminated_time",
            "last_state_loss_time",
            "last_activity_time",
            "disk_spec",
        ]:  # omit runtime attributes
            request[key] = val

    request["ssh_public_keys"] = ssh_public_keys + [sshkey]

    print_warning(
        "   => The ssh key will be added to the cluster."
        + "\n   Note: The cluster will be restarted immediately!"
    )
    answer = input(
        "   => Shall the ssh key be added and the cluster be restarted (y/n)? (default = n): "
    )
    if answer.lower() == "y":
        try:
            response = client.edit_cluster(request)
        except DatabricksApiException as ex:
            print_error(str(ex))
            return None
        print_ok("   => OK")
    else:
        print_error("   => Cancelled")


def download_notebook(url, prefix="."):
    """Download Databricks demo notebooks from docs.databricks.com.
    It adds two cells on the top to initialize the Databricks context and MLflow

    Args:
        url (str): The HTML url copied from the "Get notebook link"
    """

    def wrap(typ, cell):
        return {
            "cell_type": typ,
            "source": cell,
            "outputs": [],
            "metadata": {},
            "execution_count": 0,
        }

    #     cell1 = """# Retrieve Spark Context
    # from databrickslabs_jupyterlab.connect import dbcontext
    # from databrickslabs_jupyterlab.remote import is_remote
    # dbcontext()"""

    #     cell2 = """# Setup MLflow (optional)
    # import os
    # from mlflow
    # from mlflow.tracking import MlflowClient

    # def remote_mlflow_setup(experiment_name):
    #     mlflow.set_experiment(experiment_name)
    #     experiment = MlflowClient().get_experiment_by_name(experiment_name)
    #     from IPython.display import HTML, display
    #     url = "%s/#mlflow/experiments/%s" % (os.environ["DBJL_HOST"], experiment.experiment_id)
    #     display(HTML('View the <a href="%s">MLflow experiment</a>' % url))

    # experiment_name = <FILL>
    # home = "/Users/%s" % <FILL>
    # experiment_name = "%s/experiments/%s" % (home, experiment_name)
    # remote_mlflow_setup(experiment_name)"""

    cell = "**Downloaded by Databrickslabs Jupyterlab**"

    print("*** experimental ***")
    name = os.path.splitext(os.path.basename(url))[0]

    if url.startswith("http"):
        resp = requests.get(url)
        if resp.status_code != 200:
            print("error")
            return
        code = resp.text
    else:
        with open(url, "r") as fd:
            code = fd.read()

    r = re.compile("<script>(.+?)</script>")
    scripts = r.findall(code)
    script = [s for s in scripts if "__DATABRICKS_NOTEBOOK_MODEL" in s][0]
    r = re.compile("'(.+?)'")
    s = r.findall(script)
    nb = json.loads(unquote(base64.b64decode(s[0]).decode("utf-8")))
    language = nb["language"]
    cells = [c["command"] for c in nb["commands"]]
    ipy = {
        # "cells": [wrap("code", cell1), wrap("code", cell2)],
        "cells": [wrap("markdown", cell)],
        "nbformat": 4,
        "nbformat_minor": 4,
        "metadata": {},
    }

    for c in cells:
        t = "code"
        if c.startswith("%md"):
            t = "markdown"
            c = c[3:]
            r = re.compile(r"(#+)([^#\s])")
            c = r.sub(r"\1 \2", c)
        elif c.startswith("%python"):
            c = c[8:]
        elif c.startswith("%sql"):
            c = c.replace("%sql", "%%sql\n")
        elif c.startswith("%sh"):
            c = c.replace("%sh", "%%sh\n")
        elif c.startswith("%scala"):
            c = c.replace("%scala", "%%scala\n")
        else:
            if language != "python":
                c = "%%" + language + "\n" + c

        # SQL cells can have multiple statements separated by ";" in Databricks.
        # So split them in sigle cells for Jupyterlab Integration
        if c.startswith("%%sql"):
            # remove potential trailing ";" and then split at ";" to get all SQL statements
            lines = c[5:].strip(";").split(";")
            for line in lines:
                ipy["cells"].append(wrap(t, "%%sql\n" + line))
        else:
            ipy["cells"].append(wrap(t, c))

    with open(os.path.join(prefix, "%s.ipynb" % name), "w") as fd:
        fd.write(json.dumps(ipy))

    print("Downloaded notebook to %s.ipynb" % name)



import os
import platform
import subprocess
import sys
import tempfile
from html import escape

is_windows = platform.platform(1, 1).split("-")[0] == "Windows"


def bye(status=0):
    """Standard exit function

    Args:
        msg (str, optional): Exit message to be printed. Defaults to None.
    """
    if status != 0:
        print_error("\n=> Exiting")
    sys.exit(status)


def _print(color, *args):
    # Under Windows prompt_toolkit does not work with pytest
    # So add an environment variable check to switch to simple output
    if is_windows and (os.environ.get("PYTEST_CURRENT_TEST") is not None):
        print(*args)
    else:
        from prompt_toolkit import print_formatted_text, HTML
        from prompt_toolkit.styles import Style

        style = Style.from_dict({"error": "#ff0000", "ok": "#00ff00", "warning": "#ff00ff"})
        print_formatted_text(
            HTML(
                "<{color}>{message}</{color}>".format(color=color, message=escape(" ".join(args)))
            ),
            style=style,
        )


def print_ok(*args):
    _print("ok", *args)


def print_error(*args):
    _print("error", *args)


def print_warning(*args):
    _print("warning", *args)


def question(tag, message, choices):
    # Under Windows prompt_toolkit does not work with pytest
    # So only load questionary when this function is called - which is not used by automated tests
    import questionary
    from prompt_toolkit.styles import Style

    custom_style_fancy = Style(
        [
            ("answer", "fg:#f44336 bold"),  # submitted answer text behind the question
            (
                "highlighted",
                "fg:#f44336 bold",
            ),  # pointed-at choice in select and checkbox prompts
        ]
    )

    answer = questionary.select(
        message,
        choices=choices,
        style=custom_style_fancy,
    ).ask()  # returns value of selection
    return {tag: answer}


def utf8_decode(text):
    if isinstance(text, str):
        return text

    try:
        return text.decode("utf-8")
    except:  # pylint: disable=bare-except
        # ok, let's replace the "bad" characters
        return text.decode("utf-8", "replace")


def execute(cmd):
    """Execute subprocess

    Args:
        cmd (list(str)): Command as list of cmd parts (e.g. ["ls", "-l"])
    """
    if is_windows and cmd[0] == "conda":
        cmd[0] = "conda.bat"

    try:
        # Cannot use encoding arg at the moment, since need to support python 3.5
        result = subprocess.run(
            cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE, check=False
        ).__dict__
        result["stderr"] = utf8_decode(result["stderr"])
        result["stdout"] = utf8_decode(result["stdout"])
        # print(result["returncode"])
        # print(result["stdout"])
        # print(result["stderr"])
        return result
    except Exception as ex:  # pylint: disable=broad-except
        print("%s; %s" % (type(ex), ex))
        return {"args": cmd, "returncode": -1, "stdout": "", "stderr": str(ex)}


def execute_script(script, success_message, error_message):
    with tempfile.TemporaryDirectory() as tmpdir:
        script_name = "db_jlab_script.bat" if is_windows else "db_jlab_script.sh"

        script_file = os.path.join(tmpdir, script_name)
        with open(script_file, "w") as fd:
            fd.write(script)

        cmd = [script_file] if is_windows else ["bash", script_file]

        result = execute(cmd)
        if result["returncode"] != 0:
            print_error(error_message)
            print_error(str(result))
            sys.exit(1)
        else:
            print_ok(success_message)


# class SshConfig:
#     def __init__(self):
#         self.config_file = os.path.expanduser("~/.ssh/config")
#         self.load()

#     def _get_mtime(self):
#         return os.stat(self.config_file).st_mtime

#     def load(self):
#         self.mtime = self._get_mtime()
#         sc = ssh_config.SSHConfig.load(self.config_file)
#         self.hosts = {
#             h.name: h.attributes()["HostName"]
#             for h in sc.hosts()
#             if h.attributes().get("HostName", None) is not None
#         }

#     def get_dns(self, host_name):
#         if self._get_mtime() > self.mtime:
#             self.load()
#         return self.hosts.get(host_name, None)



import os
import re

import pandas as pd

from ipywidgets import Select, VBox, HBox, Button, Output
from IPython.display import display, HTML, Markdown, Code, Image


class Dbfs(object):
    """Database browser implementation

    Args:
        dbutils (DBUtils): DBUtils object (for fs only)
    """

    def __init__(self, dbutils):
        self.dbutils = dbutils
        self.running = False
        self.path = None
        self.flist = None
        self.refresh = None
        self.path_view = None
        self.preview = None
        self.up = None

    def create(self, path="/", height="400px"):
        if self.running:
            print("dbfs browser already running. Use close() first")
            return
        self.path = path
        self.flist = Select(options=[], disabled=False, layout={"height": height})
        self.flist.observe(self.on_click, names="value")

        self.refresh = Button(icon="refresh", layout={"width": "40px"})
        self.refresh.on_click(self.on_refresh)
        self.path_view = Output()
        self.preview = Output(
            layout={
                "width": "800px",
                "height": height,
                "overflow": "scroll",
                "border": "1px solid gray",
            }
        )

        self.up = Button(icon="arrow-up", layout={"width": "40px"})
        self.up.on_click(self.on_up)

        display(
            VBox([HBox([self.refresh, self.up, self.path_view]), HBox([self.flist, self.preview]),])
        )

        self.update()
        self.running = True

    def convertBytes(self, fsize):
        """Convert bytes to largest unit

        Args:
            fsize (int): Size in bytes

        Returns:
            tuple: size of largest unit, largest unit
        """
        size = fsize
        unit = "B"
        if size > 1024 * 1024 * 1024 * 10:
            size = int(size / 1024.0 / 1024.0 / 1024.0)
            unit = "GB"
        elif size > 1024 * 1024 * 10:
            size = int(size / 1024.0 / 1024.0)
            unit = "MB"
        elif size > 1024 * 10:
            size = int(size / 1024.0)
            unit = "KB"
        return (size, unit)

    def update(self):
        """Update the view when an element was selected"""
        self.path_view.clear_output()
        self.preview.clear_output()
        with self.path_view:
            print("updating ...")
        try:
            fobjs = self.dbutils.fs.ls(self.path)
        except:  # pylint: disable=bare-except
            with self.path_view:
                print("Error: Cannot access folder")
            return False

        self.show_path(self.path)

        dirs = sorted([fobj.name for fobj in fobjs if fobj.isDir()], key=lambda x: x.lower())
        files = sorted(
            [
                "%s (%d %s)" % ((fobj.name,) + self.convertBytes(fobj.size))
                for fobj in fobjs
                if not fobj.isDir()
            ],
            key=lambda x: x[0].lower(),
        )
        self.flist.options = [""] + dirs + files
        return True

    def show_path(self, path):
        """Show path in output widget

        Args:
            path (str): Currently selected path
        """
        self.path_view.clear_output()
        with self.path_view:
            print("dbfs:" + re.sub(r"\s\(.*?\)$", "", path))

    def show_preview(self, path):
        """Show preview of csv, md or txt in output widget

        Args:
            path (str): Currently selected path
        """
        real_path = re.sub(r"\s\(.*?\)$", "", path)
        parts = real_path.split(".")
        if len(parts) > 0 and parts[-1].lower() in [
            "md",
            "html",
            "csv",
            "txt",
            "sh",
            "sql",
            "py",
            "scala",
            "json",
            "jpg",
            "jpeg",
            "png",
            "gif",
        ]:
            ext = parts[-1].lower()
            filename = "/dbfs" + real_path
            # text = self.dbutils.fs.head(real_path)
            self.preview.clear_output()
            with self.preview:
                if ext == "html":
                    display(HTML(filename=filename))
                elif ext in ["py", "sh", "sql", "scala"]:
                    display(Code(filename=filename))
                elif ext in ["jpg", "jpeg", "png", "gif"]:
                    display(Image(filename=filename))
                elif ext == "md":
                    display(Markdown(filename=filename))
                elif ext == "csv":
                    df = pd.read_csv(filename)
                    display(df)
                else:
                    with open(filename, "r") as fd:
                        print(fd.read())

    def on_refresh(self, _):
        """Refresh handler

        Args:
            b (ipywidgets.Button): clicked button
        """
        self.update()

    def on_up(self, _):
        """Up handler

        Args:
            b (ipywidgets.Button): clicked button
        """
        new_path = os.path.dirname(self.path.rstrip("/"))
        if new_path != self.path:
            self.path = new_path
        self.update()

    def on_click(self, change):
        """Click handler providing db and parent as context

        Args:
            db (str): database name
            parent (object): parent object
        """
        new_path = os.path.join(self.path, change["new"])
        if change["old"] is not None:
            if len(change["new"]) > 0 and change["new"][-1] == "/":
                old_path = self.path
                self.path = new_path
                if not self.update():
                    self.path = old_path
            else:
                self.show_path(new_path)
                self.show_preview(new_path)

    def close(self):
        """Close view"""
        self.running = False



import configparser
import json
import os
import shutil
import sys
import time
from os.path import expanduser

from jupyter_client import kernelspec
from .ssh_config import SshConfig

from databrickslabs_jupyterlab.utils import (
    bye,
    print_ok,
    print_error,
    print_warning,
    question,
    execute,
)

PREFIX = "ssh_"


def conda_version():
    """Check conda version"""
    result = execute(["conda", "--version"])
    if result["returncode"] != 0:
        print_error(result["stderr"])
        sys.exit(1)

    if result["stdout"].startswith("conda"):
        return result["stdout"].strip().split(" ")[1]
    elif result["stderr"].startswith("conda"):
        return result["stderr"].strip().split(" ")[1]
    else:
        return None


def write_config():
    """Store jupyter lab configuration necessary for databrickslabs_jupyterlab
    Set values:
    - c.KernelManager.autorestart: False
    - c.MappingKernelManager.kernel_info_timeout: 600
    """
    config = {
        "c.KernelManager.autorestart": False,
        "c.MappingKernelManager.kernel_info_timeout": 20,
    }
    c1, c2 = list(config.keys())
    full_path = os.path.expanduser("~/.jupyter")
    if not os.path.exists(full_path):
        os.mkdir(full_path)
    config_file = os.path.join(full_path, "jupyter_notebook_config.py")
    lines = []
    if os.path.exists(config_file):
        with open(config_file, "r") as fd:
            lines = fd.read().split("\n")
        # avoid growing by one empty line for each run
        if lines[-1] == "":
            lines = lines[:-1]
    with open(config_file, "w") as fd:
        for line in lines:
            if line.startswith(c1) or line.startswith(c2):
                kv = line.split("=")
                k = kv[0].strip()
                if config.get(k) is not None:  #  if None it's a duplicate, so ignore
                    fd.write("%s = %s\n" % (k, config[k]))
                    del config[k]
            else:
                # else just copy line
                fd.write("%s\n" % line)
        # write all missing configs
        fd.write("\n".join(["%s = %s" % (k, v) for k, v in config.items()]))


def get_local_libs():
    """Get installed libraries of the currtent conda environment"""
    result = execute(["conda", "list", "--json"])
    if result["returncode"] != 0:
        print_error(result["stderr"])
        return False

    try:
        return json.loads(result["stdout"])
    except Exception as ex:  # pylint: disable=broad-except
        print_error(ex)
        return False


def get_db_config(profile):
    """Get Databricks configuration from ~/.databricks.cfg for given profile

    Args:
        profile (str): Databricks CLI profile string

    Returns:
        tuple: The tuple of host and personal access token from ~/.databrickscfg
    """
    config = configparser.ConfigParser()
    configs = config.read(expanduser("~/.databrickscfg"))
    if not configs:
        print_error("Cannot read ~/.databrickscfg")
        bye(1)

    profiles = config.sections()
    if not profile in profiles:
        print(" The profile '%s' is not available in ~/.databrickscfg:" % profile)
        for p in profiles:
            print("- %s" % p)
        bye(1)
    else:
        host = config[profile]["host"]
        token = config[profile]["token"]
        return host, token


def add_known_host(address, port, known_hosts="~/.ssh/known_hosts"):
    result = execute(["ssh-keygen", "-R", "[%s]:%s" % (address, port)])
    result = execute(["ssh-keyscan", "-p", str(port), address])
    if result["returncode"] == 0:
        fingerprint = result["stdout"]
        known_hosts_file = os.path.expanduser(known_hosts)
        with open(known_hosts_file, "a") as fd:
            fd.write("\n%s" % fingerprint)
        print("   => Known hosts fingerprint added for %s:%s\n" % (address, port))
    else:
        print_warning("   => Could not add know_hosts fingerprint for %s:%s\n" % (address, port))


def prepare_ssh_config(cluster_id, profile, endpoint):
    """Add/edit the ssh configuration belonging to the given cluster in ~/.ssh/config

    Args:
        cluster_id (str): Cluster ID
        profile (str): Databricks CLI profile string
        endpoint (str): Public DNS/IP address and Port as "address:port"
    """
    backup_path = "~/.databrickslabs_jupyterlab/ssh_config_backup"

    config = os.path.expanduser("~/.ssh/config")
    if not os.path.exists(os.path.expanduser(backup_path)):
        os.makedirs(os.path.expanduser(backup_path))

    data = ""
    if os.path.exists(config):
        print_warning("   => ~/.ssh/config will be changed")
        backup = "%s/config.%s" % (backup_path, time.strftime("%Y-%m-%d_%H-%M-%S"))
        print_warning("   => A backup of the current ~/.ssh/config has been created")
        print_warning("   => at %s" % backup)

        shutil.copy(config, os.path.expanduser(backup))

    ssh_config = SshConfig(config)

    host = ssh_config.get_host(cluster_id)
    if host is None:
        host = ssh_config.add_host(cluster_id)

    address, port = endpoint.split(":")
    host.set_param("HostName", address)
    host.set_param("IdentityFile", "~/.ssh/id_%s" % profile)
    host.set_param("Port", port)
    host.set_param("User", "ubuntu")
    host.set_param("ServerAliveInterval", 30)
    host.set_param("ServerAliveCountMax", 5760)
    host.set_param("ConnectTimeout", 5)

    print(f"   => Jupyterlab Integration made the following changes to {config}:")
    ssh_config.dump()
    ssh_config.write()

    add_known_host(address, port)


def show_profiles():
    """Show locally configured Databricks CLI profile"""
    template = "%-20s %-60s %s"
    print("")
    print(template % ("PROFILE", "HOST", "SSH KEY"))
    config = configparser.ConfigParser()
    config.read(expanduser("~/.databrickscfg"))
    profiles = config.sections()

    for profile in profiles:
        host, _ = get_db_config(profile)
        ssh_ok = "OK" if os.path.exists(os.path.expanduser("~/.ssh/id_%s") % profile) else "MISSING"
        print(template % (profile, host, ssh_ok))


def create_kernelspec(
    profile,
    organisation,
    host,
    cluster_id,
    cluster_name,
    local_env,
    python_path,
    no_spark,
    extra_env="",
):
    """Create or edit the ssh_ipykernel specification for jupyter lab

    Args:
        profile (str): Databricks CLI profile string
        organisation (str): In case of Azure, the organization ID, else None
        host (str): host from databricks cli config for given profile string
        cluster_id (str): Cluster ID
        cluster_name (str): Cluster name
        local_env (str): Name of the local conda environment
        python_path (str): Remote python path to be used for kernel
        extra_env (str): Env vars to add to the remote notebook like "V1=ABC V2=DEF"
    """
    from ssh_ipykernel.manage import add_kernel  # pylint: disable=import-outside-toplevel

    print("   => Creating kernel specification for profile '%s'" % profile)
    env = "DBJL_PROFILE=%s DBJL_HOST=%s DBJL_CLUSTER=%s" % (profile, host, cluster_id)
    if organisation is not None:
        env += " DBJL_ORG=%s" % organisation
    if extra_env != "":
        env = env + " " + extra_env

    if cluster_name.replace(" ", "_") == local_env:
        display_name = "SSH %s %s:%s" % (cluster_id, profile, cluster_name)
        if not no_spark:
            display_name += "(Spark)"
    else:
        display_name = "SSH %s %s:%s (%s%s)" % (
            cluster_id,
            profile,
            cluster_name,
            local_env,
            ("" if no_spark else "/Spark"),
        )

    add_kernel(
        host=cluster_id,
        display_name=display_name,
        local_python_path=sys.executable,
        remote_python_path=os.path.dirname(python_path),
        sudo=True,
        env=env,
        timeout=5,
        module="databrickslabs_jupyterlab",
        opt_args=["--no-spark"] if no_spark else [],
    )
    print("   => Kernel specification '%s' created or updated" % (display_name))


def remove_kernelspecs():
    km = kernelspec.KernelSpecManager()

    while True:

        remote_kernels = {
            kernelspec.get_kernel_spec(k).display_name: k
            for k, v in kernelspec.find_kernel_specs().items()
            if k.startswith("ssh_")
        }
        if remote_kernels == {}:
            print_ok("   => No databricklabs_jupyterlab kernel spec left")
            break

        answer = question(
            "kernel_name",
            "Which kernel spec to delete (Ctrl-C to finish)?",
            list(remote_kernels.keys()),
        )
        kernel_name = answer["kernel_name"]

        if kernel_name is None:
            break

        answer = input("Really delete kernels spec '%s' (y/n) " % kernel_name)
        if answer.lower() == "y":
            km.remove_kernel_spec(remote_kernels[kernel_name])



from datetime import datetime
import os
from IPython.display import HTML
import pandas as pd
from mlflow.tracking import MlflowClient


class MLflowBrowser:
    def __init__(self, experiment_name):
        self.client = MlflowClient()
        self.experiment = self.client.get_experiment_by_name(experiment_name)
        self.experiment_id = self.experiment.experiment_id
        self.url = "{host}/_mlflow/#/experiments/{experiment_id}/runs/".format(
            host=os.environ["DBJL_HOST"].strip("/"), experiment_id=self.experiment_id
        )
        self.runs = []
        self.table = None

    def _make_clickable(self, run_id, shorten=True):
        if shorten and run_id != "-":
            view_run_id = "%s..." % run_id[:8]
        else:
            view_run_id = run_id
        return '<a target="_blank" href="{}">{}</a>'.format(self.url + run_id, view_run_id)

    def _make_status(self, status):
        if status == "FINISHED":
            return u"\N{white heavy check mark}"
        elif status == "FAILED":
            return u"\N{cross mark}"
        elif status == "RUNNING":
            return u"\N{cyclone}"
        return status

    def summarize_run(self, run):
        result = dict(
            status=run.info.status,
            date=datetime.utcfromtimestamp(run.info.start_time / 1000).isoformat(
                sep=" ", timespec="seconds"
            ),
            run_id=run.info.run_id,
            parent_run_id=run.data.tags.get("mlflow.parentRunId", "-"),
            run_name=run.data.tags.get("mlflow.runName", "-"),
            user=run.data.tags.get("mlflow.user", "-"),
        )
        for k, v in run.data.params.items():
            try:
                val = float(v)
            except ValueError:
                val = v
            result["P.%s" % k] = val

        for k, v in run.data.metrics.items():
            try:
                val = float(v)
            except ValueError:
                val = v
            result["M.%s" % k] = val

        return result

    def get_runs(self, run_id=None):
        self.runs = self.client.search_runs(self.experiment_id)
        df = pd.DataFrame([self.summarize_run(run) for run in self.runs]).sort_values(
            ["date"], ascending=False
        )
        params = sorted([c for c in df.columns if c.startswith("P.")])
        metrics = sorted([c for c in df.columns if c.startswith("M.")])
        self.table = df[
            ["status", "date", "run_id", "parent_run_id", "run_name"] + params + metrics + ["user"]
        ]
        if run_id is None:
            return self.table
        else:
            return self.table[(self.table["run_id"] == run_id) | (self.table["parent_run_id"] == run_id)]

    def display(self, runs):
        print("=> Click on a run_id to open the run on the managed MLflow Tracking Server\n")

        return runs.fillna("-").style.format(
            {
                "run_id": self._make_clickable,
                "status": self._make_status,
                "parent_run_id": self._make_clickable,
            }
        )

    def compare(self, runs):
        run_ids = ",".join(["%%22%s%%22" % run for run in runs["run_id"]])
        url = "{host}/_mlflow/#/compare-runs?runs=[{run_ids}]&experiment={experiment_id}".format(
            host=os.environ["DBJL_HOST"].strip("/"),
            experiment_id=self.experiment_id,
            run_ids=run_ids,
        )
        return HTML(
            "<a href='{}'>Click to open comparision in Managed Tracking Server</a>".format(url)
        )



import os
import platform
import re
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def notebook_version():
    with open("env.yml", "r") as fd:
        env = fd.read()
    r = re.compile("notebook==.*\n")
    nb = r.search(env).group().strip().split("==")[1]
    return nb


INSTALL_REQUIRES = [
]

EXTRAS_REQUIRES = {
    "cli": [
        "notebook>=6.3.0,<6.4.0",
        "version_parser>=1.0.0,<1.1.0",
        "questionary",
        "databricks_cli",         
        "jupyterlab>=3.0.0,<3.1.0",
        "ipywidgets>=7.6.3,<7.7.0",
        "ipykernel>=5.5.3,<5.6.0",
        "tornado>=6.1.0,<6.2.0",
        "widgetsnbextension==3.5.1",
        "ssh-ipykernel==1.1.0",
        "inquirer==2.6.3",
        "pyperclip>=1.8.2<1.9.0",
    ],
    "dev": [
        "pytest", 
        "jupyter-console", 
        "pyyaml", 
        "bumpconfig", 
        "twine", 
        "yapf", 
        "pylint"
    ],
}

setup(
    name="databrickslabs_jupyterlab",
    version="2.2.0-rc0",
    author="Bernhard Walter",
    author_email="bernhard.walter@databricks.com",
    url="https://github.com/databrickslabs/Jupyterlab-Integration",
    description=("Remote JupyterLab kernel for Databricks"),
    long_description_content_type="text/markdown",
    license="Databricks License",
    keywords="databricks jupyter jupyterlab spark",
    packages=find_packages(exclude=["databrickslabs-jupyterlab-status"]),
    scripts=["databrickslabs-jupyterlab", "dj", "dj.bat"],
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRES,
    data_files=[
        (
            "etc/jupyter/jupyter_notebook_config.d",
            ["databrickslabs_jupyterlab/status/etc/serverextension.json"],
        ),
    ],
    include_package_data=True,
    long_description=read("PYPI.md"),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: Other/Proprietary License",
    ],
    zip_safe=False,
)
